/*
 * Copyright (c) 2013 Nathaniel Jones
 * Massachusetts Institute of Technology
 */

#include <optix.h>
#include <optixu/optixu_math_namespace.h>
#include "raytypes.h"

using namespace optix;

#define  FRESNE(ci)	(expf(-5.85f*(ci)) - 0.00287989916f)
#define  FRESTHRESH	0.017999f	/* minimum specularity for approx. */

/* Context variables */
rtDeclareVariable(unsigned int, radiance_ray_type, , );
rtDeclareVariable(rtObject,     top_object, , );

rtDeclareVariable(float3,       CIE_rgbf, , ); /* This is the value [ CIE_rf, CIE_gf, CIE_bf ] from color.h */

rtDeclareVariable(float,        specthresh, , ); /* This is the minimum fraction of reflection or transmission, under which no specular sampling is performed */
rtDeclareVariable(float,        specjitter, , );

rtDeclareVariable(float3,       ambval, , ); /* This is the final value used in place of an indirect light calculation */
rtDeclareVariable(int,          ambvwt, , ); /* As new indirect irradiances are computed, they will modify the default ambient value in a moving average, with the specified weight assigned to the initial value given on the command and all other weights set to 1 */
rtDeclareVariable(int,          ambounce, , ); /* Ambient bounces (ab) */
//rtDeclareVariable(int,          ambres, , ); /* Ambient resolution (ar) */
rtDeclareVariable(float,        ambacc, , ); /* Ambient accuracy (aa). This value will approximately equal the error from indirect illuminance interpolation */
rtDeclareVariable(int,          ambdiv, , ); /* Ambient divisions (ad) */
rtDeclareVariable(int,          ambssamp, , ); /* Ambient super-samples (as) */
rtDeclareVariable(float,        maxarad, , ); /* maximum ambient radius */
rtDeclareVariable(float,        minarad, , ); /* minimum ambient radius */
rtDeclareVariable(float,        avsum, , ); /* computed ambient value sum (log) */
rtDeclareVariable(unsigned int, navsum, , ); /* number of values in avsum */

rtDeclareVariable(float,        minweight, , ); /* minimum ray weight (lw) */
rtDeclareVariable(int,          maxdepth, , ); /* maximum recursion depth (lr) */

rtBuffer<BasicLight> lights;

/* Material variables */
rtDeclareVariable(unsigned int, type, , ); /* The material type representing "plastic", "metal", or "trans" */
rtDeclareVariable(float3,       color, , ); /* The material color given by the rad file "plastic", "metal", or "trans" object */
rtDeclareVariable(float,        spec, , ); /* The material specularity given by the rad file "plastic", "metal", or "trans" object */
rtDeclareVariable(float,        rough, , ); /* The material roughness given by the rad file "plastic", "metal", or "trans" object */
rtDeclareVariable(float,        trans, , ) = 0.0f; /* The material transmissivity given by the rad file "trans" object */
rtDeclareVariable(float,        tspec, , ) = 0.0f; /* The material transmitted specular component given by the rad file "trans" object */

/* OptiX variables */
rtDeclareVariable(Ray, ray, rtCurrentRay, );
rtDeclareVariable(float, t_hit, rtIntersectionDistance, );
rtDeclareVariable(PerRayData_ambient_record, prd, rtPayload, );

/* Attributes */
rtDeclareVariable(float3, geometric_normal, attribute geometric_normal, );
rtDeclareVariable(float3, shading_normal, attribute shading_normal, );


typedef struct {
	float3 v;		/* division sum (partial) */
	float  r;		/* 1/distance sum */
	float  k;		/* variance for this division */
	int    n;		/* number of subsamples */
	unsigned short  t, p;	/* theta, phi indices */
}  AMBSAMP;		/* ambient sample division */

typedef struct {
	float3 ux, uy, uz;	/* x, y and z axis directions */
	float3 acoef;		/* division contribution coefficient */
	int    ns;		/* number of super-samples */
	int    nt, np;		/* number of theta and phi directions */
}  AMBHEMI;		/* ambient sample hemisphere */

static __device__ int check_overlap( float3 hit, float3 normal );
//static __device__ float3 multambient( float3 aval, float3 normal, float3 hit );
static __device__ float doambient( float3 *rcol, float3 *pg, float3 *dg, float3 normal, float3 hit );
static __device__ __inline__ void inithemi( AMBHEMI  *hp, float3 ac, float3 normal );
static __device__ int divsample( AMBSAMP  *dp, AMBHEMI  *h, float3 hit );
//static __device__ void comperrs( AMBSAMP *da, AMBHEMI *hp );
//static __device__ int ambcmp( const void *p1, const void *p2 );
//static __device__ float2 multisamp2(float r);
//static __device__ __inline__ int ilhash(int3 d);
static __device__ __inline__ float bright( float3 rgb );

RT_PROGRAM void any_hit_ambient_glass()
{
	rtIgnoreIntersection();
}

RT_PROGRAM void closest_hit_ambient()
{
	float3 world_shading_normal   = normalize( rtTransformNormal( RT_OBJECT_TO_WORLD, shading_normal ) );
	float3 world_geometric_normal = normalize( rtTransformNormal( RT_OBJECT_TO_WORLD, geometric_normal ) );

	/* check for back side */
	// if backvis is false, create a new ray starting from the hit point (i.e., ignore this hit)
	float3 ffnormal = faceforward( world_shading_normal, -ray.direction, world_geometric_normal );
	float3 hit_point = ray.origin + t_hit * ray.direction;

	// Check that this is not covered by parent
	if ( check_overlap( hit_point, ffnormal ) )
		return;

	/* compute weight */
	//float weight = 1.0f;
	//for (int i = prd.ambient_depth; i-- > 0; ) //TODO start at i-1?
	//	weight *= AVGREFL;
	//if (prd.weight < 0.1f * weight)	/* heuristic override */
	//	weight = 1.25f * prd.weight;
	float3 acol = make_float3( AVGREFL );
	float3 pg = make_float3( 0.0f );
	float3 dg = make_float3( 0.0f );

	/* compute ambient */
	float rad = doambient(&acol, &pg, &dg, ffnormal, hit_point);
	if ( rad <= FTINY )
		return;

	acol *= 1.0f / AVGREFL;		/* undo assumed reflectance */

	//if (rn != r->ron)
	//	extambient(acol, &amb, r->rop, rn);	/* texture */

	// pass the color back up the tree
	prd.result.pos = hit_point;
	prd.result.dir = ffnormal;
	prd.result.gpos = pg;
	prd.result.gdir = dg;
	prd.result.val = acol;
	//prd.result.lvl = lvl;
	//prd.result.weight = weight;
	prd.result.rad = rad;
}

// based on sumambient from ambient.c
static __device__ int check_overlap( float3 hit, float3 normal )
{
	// Check that there is a parent
	if ( !prd.parent )
		return( 0 );

	/* Ambient radius test. */
	float3 ck0 = prd.parent->pos - hit;
	float rad = prd.parent->rad;
	float e1 = optix::dot( ck0, ck0 ) / ( rad * rad );
	float acc = 0.05f * 0.05f * 1.21f; //TODO provide value for ambacc (temporarily set to zero)
	//float acc = ambacc * ambacc * 1.21f;
	if ( e1 > acc )
		return( 0 );

	/* Direction test using closest normal. */
	float d = optix::dot( prd.parent->dir, normal );
	//if (rn != r->ron) {
	//	rn_dot = DOT(av->dir, rn);
	//	if (rn_dot > 1.0-FTINY)
	//		rn_dot = 1.0-FTINY;
	//	if (rn_dot >= d-FTINY) {
	//		d = rn_dot;
	//		rn_dot = -2.0;
	//	}
	//}
	float e2 = (1.0f - d) * prd.result.weight; //TODO this is just 1.0f
	return( e2 < 0.0f || e1 + e2 <= acc );
}

static __device__ float doambient( float3 *rcol, float3 *pg, float3 *dg, float3 nrm, float3 hit_point )
{
	float  b, d;
	AMBHEMI  hemi;
	AMBSAMP  *div;
	AMBSAMP  dnew;
	float3  acol;
	AMBSAMP  *dp;
	float  arad;
	int  divcnt;
	unsigned int  i, j;
					/* initialize hemisphere */
	inithemi(&hemi, *rcol, nrm);
	divcnt = hemi.nt * hemi.np;
					/* initialize */
	//if (pg != NULL)
	//*pg = make_float3( 0.0f );
	//if (dg != NULL)
	//*dg = make_float3( 0.0f );
	*rcol = make_float3( 0.0f );
	if (divcnt == 0)
		return(0.0f); //TODO does this change the value of rcol in the calling method?
					/* allocate super-samples */
	//if (hemi.ns > 0) {// || pg != NULL || dg != NULL) {
	//	div = (AMBSAMP *)malloc(divcnt*sizeof(AMBSAMP));
	//	//if (div == NULL) // This is 0
	//	//	error(SYSTEM, "out of memory in doambient");
	//} else
		div = NULL; // This is 0
					/* sample the divisions */
	arad = 0.0f;
	acol = make_float3( 0.0f );
	if ((dp = div) == NULL)
		dp = &dnew;
	divcnt = 0;

	/* Set-up from posgradient in ambcomp.c */
	float rprevrow[32]; //rprevrow[hemi.np]; //TODO allow larger -ad
	float bprevrow[32]; //bprevrow[hemi.np];
	float xdp = 0.0f;
	float ydp = 0.0f;

	/* Set-up from dirgradient in ambcomp.c */
	float xdd = 0.0f;
	float ydd = 0.0f;

	for (j = 0u; j < hemi.np; j++) {
		/* Set-up from posgradient in ambcomp.c */
		float mag0 = 0.0f;
		float mag1 = 0.0f;
		float lastsine = 0.0f;
		float rprev, bprev;

		/* Set-up from dirgradient in ambcomp.c */
		float mag = 0.0f;

		for (i = 0u; i < hemi.nt; i++) {
			dp->t = i; dp->p = j;
			dp->v = make_float3( 0.0f );
			dp->r = 0.0f;
			dp->n = 0;
			if (divsample(dp, &hemi, hit_point) < 0) {
				rprevrow[i] = rprev = dp->r; // Set values for posgradient to avoid NaN
				bprevrow[i] = bprev = bright(dp->v);
				if (div != NULL)
					dp++;
				continue;
			}
			arad += dp->r;
			divcnt++;
			if (div != NULL)
				dp++;
			else
				acol += dp->v;

			/* Processing from posgradient in ambcomp.c */
			b = bright(dp->v);
			if (i > 0u) {
				d = rprev;//dp[-hp->np].r;
				if ( dp->r > d ) d = dp->r;
				d *= lastsine * ( 1.0f - (float)i / hemi.nt ); /* sin(t)*cos(t)^2 */
				mag0 += d * ( b - bprev ); // bright(dp[-hp->np].v)
			}
			float nextsine = sqrtf( (float)(i+1) / hemi.nt );
			if (j > 0u) {
				d = rprevrow[i];//dp[-1].r;
				if ( dp->r > d ) d = dp->r;
				mag1 += d * ( nextsine - lastsine ) * ( b - bprevrow[i] ); // bright(dp[-1].v)
			//} else {
			//	d = dp[hp->np-1].r;
			//	if ( dp->r > d ) d = dp->r;
			//	mag1 += d * (nextsine - lastsine) * (b - bright(dp[hp->np-1].v));
			}
			lastsine = nextsine;
			rprevrow[i] = rprev = dp->r;
			bprevrow[i] = bprev = b;

			/* Processing from dirgradient in ambcomp.c */
			mag += b / sqrtf( hemi.nt / ( i + 0.5f ) - 1.0f );
		}

		/* Processing from posgradient in ambcomp.c */
		mag0 *= 2.0f * M_PIf / hemi.np;
		float phi = 2.0f * M_PIf * (float)j / hemi.np;
		float cosp = cosf(phi);
		float sinp = sinf(phi);
		xdp += mag0 * cosp - mag1 * sinp;
		ydp += mag0 * sinp + mag1 * cosp;

		/* Processing from dirgradient in ambcomp.c */
		phi = 2.0f * M_PIf * ( j + 0.5f ) / hemi.np + M_PIf * 0.5f;
		xdd += mag * cosf(phi);
		ydd += mag * sinf(phi);
	}
	if (!divcnt) {
		//if (div != NULL)
		//	free((void *)div);
		return(0.0f);		/* no samples taken */
	}
	//if (divcnt < hemi.nt*hemi.np) {
	//	//pg = dg = NULL;		/* incomplete sampling */
	//	hemi.ns = 0;
	//} else if (arad > FTINY && divcnt/arad < minarad) {
	//	hemi.ns = 0;		/* close enough */
	//} else if (hemi.ns > 0) {	/* else perform super-sampling? */
	//	comperrs(div, &hemi);			/* compute errors */
	//	//qsort(div, divcnt, sizeof(AMBSAMP), ambcmp);	/* sort divs */ TODO necessary?
	//					/* super-sample */
	//	for (i = hemi.ns; i > 0u; i--) {
	//		dnew = *div;
	//		if (divsample(&dnew, &hemi, hit_point) < 0) {
	//			dp++;
	//			continue;
	//		}
	//		dp = div;		/* reinsert */
	//		j = divcnt < i ? divcnt : i;
	//		while (--j > 0 && dnew.k < dp[1].k) {
	//			*dp = *(dp+1);
	//			dp++;
	//		}
	//		*dp = dnew;
	//	}
	//	//if (pg != NULL || dg != NULL)	/* restore order */
	//	//	qsort(div, divcnt, sizeof(AMBSAMP), ambnorm);
	//}
					/* compute returned values */
	//if (div != NULL) {
		//arad = 0.0f;		/* note: divcnt may be < nt*np */
		//for (i = hemi.nt*hemi.np, dp = div; i-- > 0u; dp++) {
		//	arad += dp->r;
		//	if (dp->n > 1) {
		//		b = 1.0f/dp->n;
		//		dp->v *= b;
		//		dp->r *= b;
		//		dp->n = 1;
		//	}
		//	acol += dp->v;
		//}
		b = bright(acol);
		if (b > FTINY) {
			b = 1.0f / b;	/* compute & normalize gradient(s) */
			//if (pg != NULL) {
				//posgradient(pg, div, &hemi);
				*pg = ( xdp * hemi.ux + ydp * hemi.uy ) * ( (hemi.nt * hemi.np) / M_PIf );
				*pg *= b;
			//}
			//if (dg != NULL) {
				//dirgradient(dg, div, &hemi);
				*dg = xdd * hemi.ux + ydd * hemi.uy;
				*dg *= b;
			//}
		}
		//free((void *)div);
	//}
	*rcol = acol;
	if (arad <= FTINY)
		arad = maxarad;
	else
		arad = (divcnt+hemi.ns)/arad;
	//if (pg != NULL) {		/* reduce radius if gradient large */
		d = dot( *pg, *pg );
		if ( d * arad * arad > 1.0f )
			arad = 1.0f / sqrtf(d);
	//}
	if (arad < minarad) {
		arad = minarad;
		if ( /*pg != NULL &&*/ d * arad * arad > 1.0f ) {	/* cap gradient */
			d = 1.0f / arad / sqrtf(d);
			*pg *= d;
		}
	}
	if ((arad /= sqrt(prd.result.weight)) > maxarad)
		arad = maxarad;
	return(arad);
}

/* initialize sampling hemisphere */
static __device__ __inline__ void inithemi( AMBHEMI  *hp, float3 ac, float3 nrm )
{
	float	d;
	int  i;
	float wt = prd.result.weight;
					/* set number of divisions */
	if (ambacc <= FTINY && wt > (d = 0.8f * fmaxf(ac) * wt / (ambdiv*minweight)))
		wt = d;			/* avoid ray termination */
	hp->nt = sqrtf(ambdiv * wt / M_PIf) + 0.5f;
	i = ambacc > FTINY ? 3 : 1;	/* minimum number of samples */
	if (hp->nt < i)
		hp->nt = i;
	hp->np = M_PIf * hp->nt + 0.5f;
					/* set number of super-samples */
	hp->ns = ambssamp * wt + 0.5f;
					/* assign coefficient */
	hp->acoef = ac;
	d = 1.0f/(hp->nt*hp->np);
	hp->acoef *= d;
					/* make axes */
	hp->uz = nrm;
	hp->uy = make_float3( 0.0f );
	if ( hp->uz.x < 0.6f && hp->uz.x >= -0.6f )
		hp->uy.x = 1.0f;
	else if ( hp->uz.y < 0.6f && hp->uz.y > -0.6f )
		hp->uy.y = 1.0f;
	else //if ( hp->uz.z < 0.6f && hp->uz.z > -0.6f )
		hp->uy.z = 1.0f;
	//else
	//	error(CONSISTENCY, "bad ray direction in inithemi");
	hp->ux = normalize( cross(hp->uy, hp->uz) );
	hp->uy = normalize( cross(hp->uz, hp->ux) );
}

/* sample a division */
static __device__ int divsample( AMBSAMP  *dp, AMBHEMI  *h, float3 hit_point )
{
	PerRayData_radiance new_prd;
	//RAY  ar;
	//float3 rcoef; /* contribution coefficient w.r.t. parent */
	//int3  hlist;
	float2  spt;
	float  xd, yd, zd;
	float  b2;
	float  phi;
					/* ambient coefficient for weight */
	if (ambacc > FTINY)
		b2 = AVGREFL; // Reusing this variable
	else
		b2 = fmaxf(h->acoef);
	new_prd.weight = prd.result.weight * b2;
	if (new_prd.weight < minweight) //if (rayorigin(&ar, AMBIENT, r, ar.rcoef) < 0)
		return(-1);
	//if (ambacc > FTINY) {
	//	rcoef *= h->acoef;
	//	rcoef *= 1.0f / AVGREFL; // This all seems unnecessary
	//}
	//hlist = make_int3( prd.seed, dp->t, dp->p );
	//multisamp(spt, 2, urand(ilhash(hlist,3)+dp->n));//TODO implement
	//spt = multisamp2( frandom() );
	//int il = ilhash( hlist );
	//spt = make_float2( rnd( il ) );
	//spt = make_float2( rnd( prd.seed ) );
	spt = make_float2( curand_uniform( prd.state ), curand_uniform( prd.state ) );
	zd = sqrtf((dp->t + spt.x)/h->nt);
	phi = 2.0f*M_PIf * (dp->p + spt.y)/h->np;
	xd = cosf(phi) * zd;
	yd = sinf(phi) * zd;
	zd = sqrtf(1.0f - zd*zd);
	float3 rdir = normalize( xd*h->ux + yd*h->uy + zd*h->uz );
	//dimlist[ndims++] = dp->t*h->np + dp->p + 90171;

	new_prd.depth = 1;//prd.depth;
	new_prd.ambient_depth = 1;//prd.ambient_depth + 1;
	//new_prd.seed = prd.seed;//lcg( prd.seed );
	new_prd.state = prd.state;
//#ifdef RAYCOUNT
//	new_prd.ray_count = 1;
//#endif
	Ray amb_ray = make_Ray( hit_point, rdir, radiance_ray_type, FTINY, RT_DEFAULT_MAX );
	rtTrace(top_object, amb_ray, new_prd);
//#ifdef RAYCOUNT
//	prd.ray_count += new_prd.ray_count;
//#endif

	//ndims--;
	new_prd.result *= h->acoef;	/* apply coefficient */
	dp->v += new_prd.result;
					/* use rt to improve gradient calc */
	if (new_prd.distance > FTINY && new_prd.distance < RT_DEFAULT_MAX)
		dp->r += 1.0f/new_prd.distance; //TODO should this be sum of distances?

					/* (re)initialize error */
	if (dp->n++) {
		b2 = bright(dp->v)/dp->n - bright(new_prd.result);
		b2 = b2*b2 + dp->k*((dp->n-1)*(dp->n-1));
		dp->k = b2/(dp->n*dp->n);
	} else
		dp->k = 0.0f;
	return(0);
}

/* compute initial error estimates */
//static __device__ void comperrs( AMBSAMP *da, AMBHEMI *hp )
//{
//	float  b, b2;
//	int  i, j;
//	AMBSAMP  *dp;
//				/* sum differences from neighbors */
//	dp = da;
//	for (i = 0; i < hp->nt; i++)
//		for (j = 0; j < hp->np; j++) {
////#ifdef  DEBUG
////			if (dp->t != i || dp->p != j)
////				error(CONSISTENCY,
////					"division order in comperrs");
////#endif
//			b = bright(dp[0].v);
//			if (i > 0) {		/* from above */
//				b2 = bright(dp[-hp->np].v) - b;
//				b2 *= b2 * 0.25f;
//				dp[0].k += b2;
//				dp[-hp->np].k += b2;
//			}
//			if (j > 0) {		/* from behind */
//				b2 = bright(dp[-1].v) - b;
//				b2 *= b2 * 0.25f;
//				dp[0].k += b2;
//				dp[-1].k += b2;
//			} else {		/* around */
//				b2 = bright(dp[hp->np-1].v) - b;
//				b2 *= b2 * 0.25f;
//				dp[0].k += b2;
//				dp[hp->np-1].k += b2;
//			}
//			dp++;
//		}
//				/* divide by number of neighbors */
//	dp = da;
//	for (j = 0; j < hp->np; j++)		/* top row */
//		(dp++)->k *= 1.0f/3.0f;
//	if (hp->nt < 2)
//		return;
//	for (i = 1; i < hp->nt-1; i++)		/* central region */
//		for (j = 0; j < hp->np; j++)
//			(dp++)->k *= 0.25f;
//	for (j = 0; j < hp->np; j++)		/* bottom row */
//		(dp++)->k *= 1.0f/3.0f;
//}

/* decreasing order */
//static __device__ int ambcmp( const void *p1, const void *p2 )
//{
//	const AMBSAMP	*d1 = (const AMBSAMP *)p1;
//	const AMBSAMP	*d2 = (const AMBSAMP *)p2;
//
//	if (d1->k < d2->k)
//		return(1);
//	if (d1->k > d2->k)
//		return(-1);
//	return(0);
//}

/* convert 1-dimensional sample to 2 dimensions, based on multisamp.c */
//static __device__ __inline__ float2 multisamp2(float r)	/* 1-dimensional sample [0,1) */
//{
//	int	j;
//	register int	k;
//	int2	ti;
//	float	s;
//
//	ti = make_int2( 0 );
//	j = 8;
//	while (j--) {
//		k = s = r*(1<<2);
//		r = s - k;
//		ti += ti + make_int2( ((k>>2) & 1), ((k>>1) & 1) );
//	}
//	ti += make_int2( frandom() );
//	ti *= 1.0f/256.0f;
//}

/* hash a set of integer values */
//static __device__ __inline__ int ilhash(int3 d)
//{
//	register int  hval;
//
//	hval = 0;
//	hval ^= d.x * 73771;
//	hval ^= d.y * 96289;
//	hval ^= d.z * 103699;
//	return(hval & 0x7fffffff);
//}

static __device__ __inline__ float bright( float3 rgb )
{
	return dot( rgb, CIE_rgbf );
}

