/*
 * Copyright (c) 2013 Nathaniel Jones
 * Massachusetts Institute of Technology
 */

#include <optix.h>
#include <optixu/optixu_math_namespace.h>
#include <optixu/optixu_matrix_namespace.h>
#include <optixu/optixu_aabb_namespace.h>
#include "raytypes.h"

using namespace optix;

rtBuffer<AmbientRecord> ambient_records;

rtDeclareVariable(float,        ambacc, , ); /* Ambient accuracy (aa). This value will approximately equal the error from indirect illuminance interpolation */
rtDeclareVariable(float,        minarad, , ); /* minimum ambient radius */

//rtDeclareVariable(float3, ambient_value, attribute ambient_value, );
//rtDeclareVariable(float, weight, attribute weight_attribute, );
//rtDeclareVariable(float, extrapolation, attribute extrapolation_attribute, );
rtDeclareVariable(optix::Ray, ray, rtCurrentRay, );
rtDeclareVariable(PerRayData_ambient, prd, rtPayload, );

static __device__ __inline__ float3 sqrtf(const float3& v);

// Ignore the intersection so that the intersection program will continue to run for all overlapping recrods.
RT_PROGRAM void ambient_record_any_hit()
{
	//prd.wsum += weight;

	//if (extrapolation > 0.0f) {
	//	prd.result += ambient_value * (extrapolation * weight);
	//}
	rtIgnoreIntersection();
}

// based on makeambient from ambient.c
RT_PROGRAM void ambient_miss()
{
	//rtThrow( RT_EXCEPTION_USER );
}

// based on sumambient from ambient.c
RT_PROGRAM void ambient_record_intersect( int primIdx )
{
	const AmbientRecord* record = &ambient_records[primIdx];

	/* Ambient level test. */
	if ( record->lvl > prd.ambient_depth )
		return;
	if ( record->weight < 0.9f * prd.weight )
		return;

	/* Ambient radius test. */
	float3 ck0 = record->pos - ray.origin; //TODO calc hit point (or use ray origin?), test that hit point is in aabb
	float rad = record->rad;
	float e1 = optix::dot( ck0, ck0 ) / ( rad * rad );
	float acc = ambacc * ambacc * 1.21f;
	if ( e1 > acc )
		return;

	/* Direction test using closest normal. */
	float d = optix::dot( record->dir, prd.surface_normal );
	//if (rn != r->ron) {
	//	rn_dot = DOT(av->dir, rn);
	//	if (rn_dot > 1.0-FTINY)
	//		rn_dot = 1.0-FTINY;
	//	if (rn_dot >= d-FTINY) {
	//		d = rn_dot;
	//		rn_dot = -2.0;
	//	}
	//}
	float e2 = (1.0f - d) * prd.weight;
	if (e2 < 0.0f)
		e2 = 0.0f;
	else if (e1 + e2 > acc)
		return;

	/* Ray behind test. */
	d = optix::dot( ck0, record->dir + prd.surface_normal );
	if (d * 0.5f > minarad * ambacc + 0.001f )
		return;

	/* Jittering final test reduces image artifacts. */
	e1 = sqrtf(e1);
	e2 = sqrtf(e2);
	float wt = e1 + e2;
	if (wt > ambacc * ( 0.9f + 0.2f * curand_uniform( prd.state ) ) )
		return;

	float t = optix::dot( ck0, prd.surface_normal ) / optix::dot( ray.direction, prd.surface_normal );

	if( rtPotentialIntersection( t ) ) {
		/* Recompute directional error using perturbed normal */
		//if (rn_dot > 0.0) {
		//	e2 = sqrtf( ( 1.0f - rn_dot ) * prd.weight);
		//	wt = e1 + e2;
		//}
		if (wt <= 1e-3f)
			wt = 1e3f;
		else
			wt = 1.0f / wt;
		prd.wsum += wt; // This assignment to the prd would take place in the any-hit program if there were one

		/* This is extambient */
		//float d = 1.0f;			/* zeroeth order */

		/* gradient due to translation */
		d = 1.0f - optix::dot( record->gpos, ck0 );

		/* gradient due to rotation */
		ck0 = optix::cross( record->dir, prd.surface_normal );
		d += optix::dot( record->gdir, ck0 );

		if (d > 0.0f) {
			// This assignment to the prd would take place in the any-hit program if there were one
			prd.result += record->val * (d * wt);
		}

		rtReportIntersection( 0 ); // There is only one material for ambient geometry group
	}
}

RT_PROGRAM void ambient_record_bounds (int primIdx, float result[6])
{
	const AmbientRecord* record = &ambient_records[primIdx];
	const float3 center = record->pos;
	const float rad = record->rad;

	optix::Aabb* aabb = (optix::Aabb*)result;

	if(rad > 0.0f && !isinf(rad)) {
		const float3 dims = rad * sqrtf( 1.0f - record->dir * record->dir ) + FTINY; // Expanding by FTINY seems to help prevent misses
		aabb->m_min = center - dims;
		aabb->m_max = center + dims;
	} else {
		aabb->invalidate();
	}
}

/* floor */
static __device__ __inline__ float3 sqrtf(const float3& v)
{
	return make_float3(sqrtf(v.x), sqrtf(v.y), sqrtf(v.z));
}

